unit Shaders;

interface

const
  VS_Water: String = "
    uniform mat4 WVP;
    uniform mat4 W;
    uniform mat4 TexProj;
    uniform vec3 CamPos;
    uniform vec4 WaterOffset;
    uniform mat4 TexCoordTransform;
    attribute vec3 a_Position0;
    attribute vec3 a_Normal0;
    attribute vec2 a_TexCoord0;
    attribute vec2 a_TexCoord1;
    varying vec4 v_TexCoord;
    varying vec2 v_TexCoordLight;
    varying vec4 v_TexCoordProj;
    varying vec3 v_PosCamDir;
    void main () {
      gl_Position = vec4(a_Position0, 1) * WVP;
      v_TexCoord.xy = a_TexCoord0 + WaterOffset.xy;
      v_TexCoord.zw = (vec3(a_TexCoord0, 0) * mat3(TexCoordTransform)).xy + WaterOffset.zw;
      v_TexCoordLight = a_TexCoord1;
      v_TexCoordProj = vec4(a_Position0, 1) * TexProj;
      v_PosCamDir = (vec4(a_Position0, 1) * W).xyz - CamPos;
    }
  ";
  PS_Water: String = "
    #ifdef GL_ES
    precision mediump float;
    #endif
    uniform vec2 DepthClamp;
    uniform samplerCube Tex0;
    uniform sampler2D Tex1;
    uniform sampler2D Tex2;
    uniform sampler2D Tex3;
    uniform sampler2D TexLight;
    varying vec4 v_TexCoord;
    varying vec2 v_TexCoordLight;
    varying vec4 v_TexCoordProj;
    varying vec3 v_PosCamDir;
    float Unpack(const in vec4 Value) {
      return dot(Value, vec4(1.0 / 160581375.0, 1.0 / 65025.0, 1.0 / 256.0, 1.0));
    }
    void main () {
      float dm = Unpack(texture2DProj(Tex3, v_TexCoordProj));
      vec3 nm0 = texture2D(Tex1, v_TexCoord.xy).xzy * 2.0 - 1.0;
      vec3 nm1 = texture2D(Tex1, v_TexCoord.zw).xzy * 2.0 - 1.0;
      vec3 ref = normalize((nm0 + nm1) * vec3(1, 5, 1));
      vec2 TexProjShift = ref.xz * 2.0;//normalize(nm0 + nm1) * 0.2;
      vec4 TexProj = v_TexCoordProj;
      TexProj.xy += TexProjShift;
      vec4 proj = texture2DProj(Tex2, TexProj);
      float g = length(v_PosCamDir);
      float dg = clamp((g - DepthClamp.x) / (DepthClamp.y - DepthClamp.x), 0.0, 1.0);
      float d = (dm - dg);
      g = 1.0 - clamp((g - 40.0) / (100.0 - 40.0), 0.0, 1.0);
      d = clamp(d / 0.001, 0.0, 1.0);
      vec3 PosCamDirNorm = normalize(v_PosCamDir);
      ref = reflect(PosCamDirNorm, ref);
      vec4 Col0 = textureCube(Tex0, ref);
      vec4 Light = texture2D(TexLight, v_TexCoordLight);
      vec4 OutColor = vec4((proj * Light + g * Col0 * pow((Col0.x + Col0.y + Col0.z) * 0.33, 8.0)).xyz, d);
      gl_FragColor = OutColor;
    }
  ";

  VS_Sky: String = "
    uniform mat4 WVP;
    attribute vec3 a_Position0;
    varying vec3 v_TexCoord;
    void main () {
      gl_Position = vec4(a_Position0, 1) * WVP;
      v_TexCoord = normalize(a_Position0);
    }
  ";
  PS_Sky: String = "
    #ifdef GL_ES
    precision mediump float;
    #endif
    uniform samplerCube Tex0;
    varying vec3 v_TexCoord;
    void main () {
      gl_FragColor = textureCube(Tex0, v_TexCoord);
    }
  ";

  PS_Depth: String = "
    #ifdef GL_ES
    precision mediump float;
    #endif
    uniform vec2 DepthClamp;
    varying vec3 v_DepthVec;
    vec4 Pack(const in float Value) {
      const vec4 bit_shift = vec4(160581375.0, 65025.0, 256.0, 1.0);
      const vec4 bit_mask  = vec4(0.0, 1.0 / 255.0, 1.0 / 255.0, 1.0 / 255.0);
      vec4 res = fract(Value * bit_shift);
      res -= res.xxyz * bit_mask;
      return res;
    }
    void main () {
      float Depth = length(v_DepthVec);
      gl_FragColor = Pack(clamp((Depth - DepthClamp.x) / (DepthClamp.y - DepthClamp.x), 0.0, 1.0));
    }
  ";

function VS_Scene(const Bones, Textures: Integer; const Clip, LightMap: Boolean): String;
function PS_Scene(const Textures: Integer; const Clip, LightMap: Boolean): String;

function VS_Depth(const Bones: Integer): String;

implementation

function VS_Scene(const Bones, Textures: Integer; const Clip, LightMap: Boolean): String;
  var i: Integer;
begin
  Result := "
    uniform mat4 WVP;
  ";
  if Clip then
  Result := Result + "
    uniform mat4 W;
  ";
  if Bones > 0 then
  begin
    Result := Result + "
      const int MAX_BONE_COUNT = 60;
      uniform mat4 SkinPallete[MAX_BONE_COUNT];
    ";
  end;
  Result := Result + "
    attribute vec3 a_Position0;
    attribute vec2 a_TexCoord0;
  ";
  if LightMap then
  Result := Result + "
    attribute vec2 a_TexCoord1;
  ";
  if Textures > 0 then
  begin
    Result := Result + "
      attribute vec4 a_Color0;
    ";
  end;
  if Bones > 0 then
  begin
    if Bones = 1 then
    Result := Result + "
      attribute float a_BlendIndex0;
    "
    else
    Result := Result + "
      attribute vec" + IntToStr(Bones) + " a_BlendIndex0;
      attribute vec" + IntToStr(Bones) + " a_BlendWeight0;
    ";
  end;
  if LightMap then
  Result := Result + "
    varying vec4 v_TexCoord;
  "
  else
  Result := Result + "
    varying vec2 v_TexCoord;
  ";
  if Clip then
  Result := Result + "
    varying vec3 v_PosW;
  ";
  if Textures > 0 then
  begin
    Result := Result + "
      varying vec" + IntToStr(Textures + 1) + " v_TexBlend;
    ";
  end;
  Result := Result + "
    void main () {
  ";
  if Bones > 0 then
  begin
    if Bones = 1 then
    Result := Result + "
      mat4 FinalMatrix = SkinPallete[int(a_BlendIndex0)];
    "
    else
    begin
      Result := Result + "
        mat4 FinalMatrix = (
      ";
      for i := 0 to Bones - 1 do
      begin
        Result := Result + "
          SkinPallete[int(a_BlendIndex0[" + IntToStr(i) + "])] * a_BlendWeight0[" + IntToStr(i) + "]
        ";
        if i < Bones - 1 then Result := Result + '+';
      end;
      Result := Result + "
        );
      ";
    end;
    Result := Result + "
      vec4 SkinPos = vec4(a_Position0, 1) * FinalMatrix;
      gl_Position = SkinPos * WVP;
    ";
    if Clip then
    Result := Result + "
      v_PosW = (SkinPos * W).xyz;
    ";
  end
  else
  begin
    Result := Result + "
      gl_Position = vec4(a_Position0, 1) * WVP;
    ";
    if Clip then
    Result := Result + "
      v_PosW = (vec4(a_Position0, 1) * W).xyz;
    ";
  end;
  if LightMap then
  begin
    Result := Result + "
      v_TexCoord = vec4(a_TexCoord0, a_TexCoord1);
    ";
  end
  else
  begin
    Result := Result + "
      v_TexCoord = a_TexCoord0;
    "
  end;
  if Textures > 0 then
  begin
    case Textures of
      1:
      begin
        Result := Result + "
          float InvR = 1.0 - a_Color0.x;
          v_TexBlend = vec2(InvR, a_Color0.x);
        ";
      end;
      2:
      begin
        Result := Result + "
          float InvR = 1.0 - a_Color0.x;
          float InvG = 1.0 - a_Color0.y;
          v_TexBlend = vec3(InvR * InvG, InvG * a_Color0.x, a_Color0.y);
        ";
      end;
      3:
      begin
        Result := Result + "
          float InvR = 1.0 - a_Color0.x;
          float InvG = 1.0 - a_Color0.y;
          float InvB = 1.0 - a_Color0.z;
          v_TexBlend = vec4(InvR * InvG * InvB, InvG * InvB * a_Color0.x, InvB * a_Color0.y, a_Color0.z);
        ";
      end;
    end;
  end;
  Result := Result + "
    }
  ";
end;

function PS_Scene(const Textures: Integer; const Clip, LightMap: Boolean): String;
  var i: Integer;
begin
  Result := "
    #ifdef GL_ES
    precision mediump float;
    #endif
  ";
  for i := 0 to Textures do
  Result := Result + "
    uniform sampler2D Tex" + IntToStr(i) + ";
  ";
  if LightMap then
  Result := Result + "
    uniform sampler2D TexLight;
  ";
  Result := Result + "
    uniform vec4 LightAmbient;
  ";
  if Clip then
  Result := Result + "
    uniform vec4 ClipPlane;
  ";
  if LightMap then
  Result := Result + "
    varying vec4 v_TexCoord;
  "
  else
  Result := Result + "
    varying vec2 v_TexCoord;
  ";
  if Clip then
  Result := Result + "
    varying vec3 v_PosW;
  ";
  if Textures > 0 then
  Result := Result + "
    varying vec" + IntToStr(Textures + 1) + " v_TexBlend;
  ";
  Result := Result + "
    void main () {
  ";
  if Clip then
  Result := Result + "
    if (dot(v_PosW, ClipPlane.xyz) < ClipPlane.w) discard;
  ";
  for i := 0 to Textures do
  Result := Result + "
    vec4 t" + IntToStr(i) + " = texture2D(Tex" + IntToStr(i) + ", v_TexCoord.xy);
  ";
  if Textures > 0 then
  begin
    Result := Result + 'gl_FragColor = (';
    for i := 0 to Textures do
    begin
      Result := Result + 't' + IntToStr(i) + ' * v_TexBlend[' + IntToStr(i) + ']';
      if i < Textures then
      Result := Result + ' + ';
    end;
    Result := Result + ')';
  end
  else
  Result := Result + 'gl_FragColor = t0';
  if LightMap then
  Result := Result + ' * (texture2D(TexLight, v_TexCoord.zw) + LightAmbient);'
  else
  Result := Result + ' * LightAmbient;';
  Result := Result + "
    }
  ";
end;

function VS_Depth(const Bones: Integer): String;
  var i: Integer;
begin
  Result := '';
  if Bones > 0 then
  Result := Result + "
    const int MAX_BONE_COUNT = 60;
  ";
  Result := Result + "
    uniform mat4 WVP;
    uniform mat4 W;
    uniform vec3 CamPos;
  ";
  if Bones > 0 then
  Result := Result + "
    uniform mat4 SkinPallete[MAX_BONE_COUNT];
  ";
  if Bones > 0 then
  begin
    if Bones = 1 then
    Result := Result + "
      attribute float a_BlendIndex0;
    "
    else
    Result := Result + "
      attribute vec" + IntToStr(Bones) + " a_BlendIndex0;
      attribute vec" + IntToStr(Bones) + " a_BlendWeight0;
    ";
  end;
  Result := Result + "
    attribute vec3 a_Position0;
    varying vec3 v_DepthVec;
    void main () {
  ";
  if Bones > 0 then
  begin
    if Bones = 1 then
    Result := Result + "
      mat4 FinalMatrix = SkinPallete[int(a_BlendIndex0)];
    "
    else
    begin
      Result := Result + "
        mat4 FinalMatrix = (
      ";
      for i := 0 to Bones - 1 do
      begin
        Result := Result + "
          SkinPallete[int(a_BlendIndex0[" + IntToStr(i) + "])] * a_BlendWeight0[" + IntToStr(i) + "]
        ";
        if i < Bones - 1 then
        Result := Result + '+';
      end;
      Result := Result + "
        );
      ";
    end;
    Result := Result + "
      vec4 SkinPos = vec4(a_Position0, 1) * FinalMatrix;
      gl_Position = SkinPos * WVP;
    ";
  end
  else
  Result := Result + "
    gl_Position = vec4(a_Position0, 1) * WVP;
  ";
  Result := Result + "
      v_DepthVec = (vec4(a_Position0, 1) * W).xyz - CamPos;
    }
  ";
end;

end.
